//////////////////////////////////////////////////////////////////////////////
//
// Project     : a DOS full-screen editor using the object-oriented
//               facilities of C++.
//
// File name   : mark.h
//
// Classes     : Marks
//
// Description : this class holds the information required for the 'marked'
//               lines available for copying/moving/deleting.  These lines
//               can be moved between files, so there is one mark per
//               editor, not per file.     
//
// Copyright   : (c) 1991, 1993 Ian Craggs
//
//////////////////////////////////////////////////////////////////////////////
          
#include "Mark.h"

#include "Mwindow.h"


static void MarkLines(Lines* First, Lines* Last, bool isMarked)
{
  Lines* CurLine = First;

  while (CurLine != Last)
  {
    CurLine->isMarked = isMarked;
    CurLine = CurLine->NextOf();
  } // endwhile

  Last->isMarked = isMarked;
} // end MarkLines


void Marks::Reset()
{
  theNoOfLines = 0;
  theStartLine = theEndLine = NULL;
  theMarkedFileW = NULL;
  theStartCol = theEndCol = 0;
  theStartRow = theEndRow = 0;
} // end Reset


void Marks::Set(Cursors& aCursor, FileWindows& aFileW)
{
  if (theMarkedFileW != NULL && theMarkedFileW != &aFileW)
  {
    theMsgWind.setMsg("Mark already set in another file.");
    return;
  } // endif

  if ((aCursor.CurLineOf())->isReadOnly)
  {
    theMsgWind.setMsg("Can\'t mark this line.");
    return;
  } // endif

  if (theNoOfLines == 0)
  {
    theMarkedFileW = &aFileW;
    theStartLine = theEndLine = aCursor.CurLineOf();
    theStartRow = theEndRow = aCursor.CurRowOf();
    theStartLine->isMarked = TRUE;
  }
  else
  {
    int NewRow = aCursor.CurRowOf();

    if (NewRow == theStartRow || NewRow == theEndRow)
      return;

    Lines* NewLine = aCursor.CurLineOf();

    if (NewRow > theStartRow)
    {
      if (NewRow < theEndRow)
	MarkLines(NewLine->NextOf(), theEndLine, FALSE);
      else
	MarkLines(theEndLine->NextOf(), NewLine, TRUE);
      theEndLine = NewLine;
      theEndRow  = NewRow;
    }
    else
    {
      MarkLines(NewLine, theStartLine->PrevOf(), TRUE);
      theStartLine = NewLine;
      theStartRow  = NewRow;
    } // endif

  } // endif

  Lines* CurLine = theStartLine;
  int count = 0;

  while (CurLine != theEndLine)
  {
    ++count;
    CurLine = CurLine->NextOf();
  } // endwhile

  theNoOfLines = ++count;

} // end Set


void Marks::Clear()
{
  if (theNoOfLines == 0)
    return;

  MarkLines(theStartLine, theEndLine, FALSE);

  Reset();
} // end Clear


void Marks::Delete(bool EraseLines)
{
  if (theNoOfLines == 0)
    return;

  Lines* Before = theStartLine->PrevOf();
  Lines* After  = theEndLine->NextOf();

  Before->SetNext(After);              // Splice the lines surrounding
  After->SetPrev(Before);              // the mark together.

  Cursors* MarkedCursor = &(theMarkedFileW->theCursor);

  if (theMarkedFileW->theBaseRow->isMarked)       // Base row is marked.
  {
    // If cursor is in marked area, then

    if (MarkedCursor->CurLineOf()->isMarked)
    {
      // Move the cursor to first line.

      MarkedCursor->setRow(1);
      MarkedCursor->setCurRow(theStartRow);
      MarkedCursor->setCurLine(After);
    }
    else
    {
      // Move cursor up (on screen) no of marked lines on screen
      // to keep it on same file line.
      int count = 0;
      Lines *CurLine = theMarkedFileW->theBaseRow;
      while (CurLine->isMarked)
      {
	++count;
	CurLine = CurLine->NextOf();
      } // endwhile

      MarkedCursor->setRow(MarkedCursor->RowOf() - count);
      MarkedCursor->setCurRow(MarkedCursor->CurRowOf() - theNoOfLines);
      //CurLine set by reset cursor.
    } // endif

    // Move BaseRow to line after mark.
    theMarkedFileW->theBaseRow = After;
  }
  else if (MarkedCursor->CurLineOf()->isMarked)   // Base row not marked.
  {
    int diff = MarkedCursor->CurRowOf() - theStartRow;

    MarkedCursor->setRow(MarkedCursor->RowOf() - diff);

    MarkedCursor->setCurLine(After);
    MarkedCursor->setCurRow(theStartRow);

  } // endif

  theMarkedFileW->theEditBuffer.IncrChanges();

  if (EraseLines)
  {
    // Delete those ripped out lines!

    Lines* CurLine = theStartLine;

    while (CurLine != theEndLine)
    {
      Lines* Next = CurLine->NextOf();
      delete CurLine;
      CurLine = Next;
    } // endwhile

    delete theEndLine;

    Reset();                           // Clear the mark data.

  } // endif

} // end Delete


void Marks::Copy(FileWindows& ToFile, bool CopyLines)
{
  if (theNoOfLines == 0)
    return;

  Lines* CurLine = theStartLine;       // Point to start of marked block.
  Lines* ToStart;
  Lines* CurToLine;

  if (CopyLines == FALSE)
  {
    ToStart = theStartLine;
    CurToLine = theEndLine;
  }
  else
  {
    ToStart =
	new Lines(CurLine->DataOf());  // Point to start of new block.

    CurToLine = ToStart;

    while (CurLine != theEndLine)      // Make copies of marked lines.
    {
      CurLine = CurLine->NextOf();

      Lines* New = new Lines(CurLine->DataOf());

      CurToLine->SetNext(New);
      New->SetPrev(CurToLine);
      CurToLine = New;
    } // endwhile

  } // endif

  CurLine = ToFile.theCursor.CurLineOf();  // Point to line to insert at.

  Lines* NextLine = CurLine->NextOf();

  CurLine->SetNext(ToStart);
  ToStart->SetPrev(CurLine);

  CurToLine->SetNext(NextLine);
  NextLine->SetPrev(CurToLine);

  ToFile.theEditBuffer.IncrChanges();  // Indicate file has changed.

  // Add theNoOfLines to the line count in ToFile.

  // Leave mark alone for next copy.

} // end Copy


void Marks::Move(FileWindows& ToFile)
{
  if (theNoOfLines == 0)
    return;

  Delete(FALSE);                       // Take out the lines, but don't erase.
  Copy(ToFile, FALSE);                 // Copy the removed lines.
} // end Move


void Marks::Shift(direction aDir)
{
  if (theNoOfLines == 0)
    return;

  Lines* CurLine = theStartLine;

  while (CurLine != theEndLine)
  {
    Lines* Next = CurLine->NextOf();

    if (aDir == RIGHT)
      CurLine->InsertChar(' ', 0);
    else
      CurLine->DeleteChar(0);

    CurLine = Next;
  } // endwhile

  if (aDir == RIGHT)
    CurLine->InsertChar(' ', 0);
  else
    CurLine->DeleteChar(0);

} // end Shift
